from openpyxl.utils import get_column_letter, range_boundaries
import re
from typing import List, Tuple, Dict

# Regex to detect cell references like A1, B2, or ranges like A1:B2
CELL_REF_REGEX = r"('?[A-Za-z0-9_\-\[\] ]+'?![A-Z]{1,3}[0-9]+(:[A-Z]{1,3}[0-9]+)?)|([A-Z]{1,3}[0-9]+(:[A-Z]{1,3}[0-9]+)?)"  # noqa


def expand_range(range_reference: str) -> List[str]:
    """
    Expand a range reference (e.g., 'A1:A3') into a list of individual cell references.

    Args:
        range_ref (str): The range reference to expand.

    Returns:
        List[str]: A list of individual cell references.
    """
    # if there is a sheet name in the range reference, put it away for now
    if "!" in range_reference:
        sheet_name, range_reference = range_reference.split("!")
    else:
        sheet_name = None

    min_col, min_row, max_col, max_row = range_boundaries(range_reference)
    expanded_cells = []

    # Loop over rows and columns in the range
    for row in range(min_row, max_row + 1):
        for col in range(min_col, max_col + 1):
            cell_ref = f"{get_column_letter(col)}{row}"
            if sheet_name:
                expanded_cells.append(f"{sheet_name}!{cell_ref}")
            else:
                expanded_cells.append(cell_ref)

    return expanded_cells


def extract_references(formula: str) -> Tuple[List[str], List[str], Dict[str, str]]:
    """
    Extract all referenced cells and ranges from a formula using regular expressions.
    This returns a list of both individual cells and range references.

    Args:
        formula (str): The formula to extract references from.

    Returns:
        Tuple[List[str], List[str], Dict[str, str]]: A tuple containing lists of direct references,
                                                     range references, and a dictionary of dependencies.
    """
    formula = formula.replace("$", "")
    matches = re.findall(CELL_REF_REGEX, formula)
    references = [match[0] if match[0] else match[2] for match in matches]

    # Trim the extracted references
    references = [ref.strip() for ref in references]

    expanded_references = []
    dependencies = {}
    direct_references = []
    range_references = []

    for reference in references:
        if ":" in reference:  # it's a range like A1:A3
            expanded_cells = expand_range(reference)
            expanded_references.extend(expanded_cells)
            range_references.append(reference)
            # Store the range-to-cells relationship
            for cell in expanded_cells:
                dependencies[cell] = reference
        else:  # single cell
            direct_references.append(reference)

    return direct_references, range_references, dependencies



from collections import Counter


def print_summary(graph, functionsdict):
    """
    Summarize a networkx DiGraph representing a dependency graph and print the most used functions in the formulas.
    """
    strpadsize = 28
    numpadsize = 5

    print_basic_info(graph, strpadsize, numpadsize)
    print_highest_degree_nodes(graph, strpadsize, numpadsize)
    print_most_used_functions(functionsdict, strpadsize, numpadsize)


def print_basic_info(graph, strpadsize, numpadsize):
    print("===  Dependency Graph Summary ===")
    print(
        "Cell/Node count".ljust(strpadsize, " ")
        + str(graph.number_of_nodes()).rjust(numpadsize, " ")
    )
    print(
        "Dependency count".ljust(strpadsize, " ")
        + str(graph.number_of_edges()).rjust(numpadsize, " ")
    )
    print()


def print_highest_degree_nodes(graph, strpadsize, numpadsize):
    print("\n===  Most connected nodes     ===")
    degree_view = graph.degree()
    degree_counts = Counter(dict(degree_view))
    max_degree_node = degree_counts.most_common(10)

    for node, degree in max_degree_node:
        print(f"{node.ljust(strpadsize)}{str(degree).rjust(numpadsize, ' ')} ")


def print_most_used_functions(functionsdict, strpadsize, numpadsize):
    print("\n===  Most used functions      ===")
    sorted_functions = dict(
        sorted(functionsdict.items(), key=lambda item: item[1], reverse=True)
    )

    for function, count in sorted_functions.items():
        print(f"{function.ljust(strpadsize, ' ')}{str(count).rjust(numpadsize, ' ')}")



"""
This script extracts formulas from an Excel file and builds a dependency graph.
"""

from typing import List, Dict
from openpyxl import load_workbook
import networkx as nx
import re
import sys
from .graph_visualizer import visualize_dependency_graph
from .graph_summarizer import print_summary
from .excel_parser import extract_references

# Dictionary that stores the unique functions used in the formulas
# The key will be the function name and the value will be the number of times it was used
functions_dict: Dict[str, int] = {}


def log(msg: str) -> None:
    """
    Log a message to the console if verbosity is enabled using the --verbose flag.
    """
    if "--verbose" in sys.argv:
        print(msg)


def sanitize_sheetname(sheetname: str) -> str:
    """
    Remove any special characters from the sheet name.
    """
    return sheetname.replace("'", "")


def sanitize_range(rangestring: str) -> str:
    """
    Remove any special characters from the range.
    """
    if "!" in rangestring:
        sheet, range_ = rangestring.split("!")
        sheet = sheet.replace("'", "")
        return f"{sheet}!{range_}"
    return rangestring


def stat_functions(cellvalue: str) -> None:
    """
    Extract the functions used in the formula and store them in a dictionary.
    This will be used to print the most used functions in the formulas.
    """
    cellfuncs = re.findall(r"[A-Z]+\(", cellvalue)
    log(f"  Functions used: {functions_dict}")
    for function in cellfuncs:
        function = function[:-1]  # Remove the "(" from the function name
        functions_dict[function] = functions_dict.get(function, 0) + 1


def add_node(graph: nx.DiGraph, node: str, sheet: str) -> None:
    """
    Add a node to the graph with the specified sheet name.
    """
    log(f"Adding node: {node} in sheet: {sheet}")
    sheet = sanitize_sheetname(sheet)
    graph.add_node(node, sheet=sheet)


def extract_formulas_and_build_dependencies(file_path: str) -> nx.DiGraph:
    """
    Extract formulas from an Excel file and build a dependency graph.
    """
    try:
        wb = load_workbook(file_path, data_only=False)
    except Exception as e:
        log(f"Error loading workbook: {e}")
        sys.exit(1)

    graph = nx.DiGraph()

    for sheet_name in wb.sheetnames:
        ws = wb[sheet_name]
        log(f"========== Analyzing sheet: {sheet_name} ==========")
        sanitized_sheet_name = sanitize_sheetname(sheet_name)
        process_sheet(ws, sanitized_sheet_name, graph)

    return graph


def process_sheet(ws, sheet_name: str, graph: nx.DiGraph) -> None:
    """
    Process a sheet and add references to the graph.
    """
    for row in ws.iter_rows():
        for cell in row:
            if isinstance(cell.value, str) and cell.value.startswith("="):
                process_formula_cell(cell, sheet_name, graph)


def process_formula_cell(cell, sheet_name: str, graph: nx.DiGraph) -> None:
    """
    Process a cell containing a formula.
    """
    stat_functions(cell.value)
    cell_reference = f"{sheet_name}!{cell.coordinate}"
    log(f"Formula in {cell_reference}: {cell.value}")
    add_node(graph, cell_reference, sheet_name)

    direct_references, range_references, range_dependencies = extract_references(
        cell.value
    )
    add_references_to_graph(direct_references, cell_reference, sheet_name, graph)
    add_ranges_to_graph(range_references, cell_reference, sheet_name, graph)
    add_range_dependencies_to_graph(range_dependencies, sheet_name, graph)


def add_references_to_graph(
    references: List[str], current_cell: str, sheet_name: str, graph: nx.DiGraph
) -> None:
    """
    Add direct cell references to the graph.
    """
    for cell_reference in references:
        cell_reference = format_reference(cell_reference, sheet_name)
        log(f"  Cell: {cell_reference}")
        add_node(graph, cell_reference, sheet_name)
        graph.add_edge(current_cell, cell_reference)


def add_ranges_to_graph(
    ranges: List[str], current_cell: str, sheet_name: str, graph: nx.DiGraph
) -> None:
    """
    Add range references to the graph.
    """
    for range_reference in ranges:
        range_sheet_name = get_range_sheet_name(range_reference, sheet_name)
        range_reference = format_reference(range_reference, sheet_name)
        log(f"  Range: {range_reference}")
        add_node(graph, range_reference, range_sheet_name)
        graph.add_edge(current_cell, range_reference)


def add_range_dependencies_to_graph(
    range_dependencies: Dict[str, str], sheet_name: str, graph: nx.DiGraph
) -> None:
    """
    Add dependencies between ranges and cells.
    """
    for cell_reference, range_reference in range_dependencies.items():
        range_reference = format_reference(range_reference, sheet_name)
        cell_reference = format_reference(cell_reference, sheet_name)
        range_sheet_name = range_reference.split("!")[0]
        cell_sheet_name = cell_reference.split("!")[0]

        add_node(graph, cell_reference, cell_sheet_name)
        add_node(graph, range_reference, range_sheet_name)
        graph.add_edge(range_reference, cell_reference)


def format_reference(reference: str, sheet_name: str) -> str:
    """
    Format a cell or range reference to include the sheet name if not already present.
    """
    return (
        f"{sheet_name}!{reference}"
        if "!" not in reference
        else reference.replace("'", "")
    )


def get_range_sheet_name(range_reference: str, sheet_name: str) -> str:
    """
    Get the sheet name for a range reference.
    """
    return sheet_name if "!" not in range_reference else range_reference.split("!")[0]


if __name__ == "__main__":
    path_to_excel = "Book1.xlsx"

    if len(sys.argv) > 1:
        path_to_excel = sys.argv[1]

    # Extract formulas and build the dependency graph
    dependency_graph = extract_formulas_and_build_dependencies(path_to_excel)

    print_summary(dependency_graph, functions_dict)

    # if --no-visualize is not passed as argument
    if "--no-visualize" not in sys.argv:
        print(
            "\033[1;30;40m\nVisualizing the graph of dependencies.\nThis might take a while...\033[0;37;40m\n"  # noqa
        )

        visualize_dependency_graph(dependency_graph, path_to_excel)



import sys
from .graphbuilder import extract_formulas_and_build_dependencies
from .graph_summarizer import print_summary
from .graph_visualizer import visualize_dependency_graph

functions_dict = {}

def main(path_to_excel: str = "Book1.xlsx", visualize: bool = True) -> None:
    """
    Main function to extract formulas, build the dependency graph, and optionally visualize it.

    Args:
        path_to_excel (str): Path to the Excel file.
        visualize (bool): Whether to visualize the dependency graph.
    """
    # Extract formulas and build the dependency graph
    dependency_graph = extract_formulas_and_build_dependencies(path_to_excel)

    print_summary(dependency_graph, functions_dict)

    if visualize:
        print(
            "\033[1;30;40m\nVisualizing the graph of dependencies.\nThis might take a while...\033[0;37;40m\n"  # noqa
        )
        visualize_dependency_graph(dependency_graph, path_to_excel)

def run_from_command_line() -> None:
    """
    Run the main function with command-line arguments.
    """
    path_to_excel = "Book1.xlsx"
    visualize = "--no-visualize" not in sys.argv

    if len(sys.argv) > 1:
        path_to_excel = sys.argv[1]

    main(path_to_excel, visualize)

if __name__ == "__main__":
    run_from_command_line()


import matplotlib.cm as cm
import matplotlib.patches as mpatches
import networkx as nx
import matplotlib.pyplot as plt
import sys

small_graph_settings = {
    "node_size": 50,
    "edge_color": "black",
    "with_labels": False,
    "font_size": 10,
    "linewidths": 0.8,
    "alpha": 0.8,
    "width": 0.2,
}

medium_graph_settings = {
    "node_size": 30,
    "edge_color": "gray",
    "with_labels": True,
    "font_size": 10,
    "linewidths": 0.1,
    "alpha": 0.4,
    "width": 0.2,
}

large_graph_settings = {
    "node_size": 5,
    "edge_color": "gray",
    "with_labels": False,
    "font_size": 12,
    "linewidths": 0.5,
    "alpha": 0.2,
    "width": 0.2,
}


def get_graph_default_settings(graph_size):
    """
    Set the default settings for the graph visualization based on the number of nodes.
    """

    if graph_size < 200:
        settings = small_graph_settings
        fig_size = 10
    elif graph_size < 500:
        settings = medium_graph_settings
        fig_size = 20
    else:
        settings = large_graph_settings
        fig_size = 20

    return settings, fig_size


# Function to get colors and generate legend for sheets
def get_node_colors_and_legend(graph):
    sheets = {data.get("sheet", "Sheet1") for _, data in graph.nodes(data=True)}
    color_map = cm.get_cmap("tab20c", len(sheets))

    # Map sheet names to colors
    sheet_to_color = {sheet: color_map(i) for i, sheet in enumerate(sheets)}

    # Assign colors to nodes based on their sheet
    node_colors = [
        sheet_to_color[data.get("sheet", "Sheet1")]
        for _, data in graph.nodes(data=True)
    ]

    # Create patches for the legend
    legend_patches = [
        mpatches.Patch(color=color, label=sheet)
        for sheet, color in sheet_to_color.items()
    ]

    return node_colors, legend_patches


def visualize_dependency_graph(graph, file_path):
    """
    Render the dependency graph using matplotlib and networkx.
    """

    if "--keep-direction" not in sys.argv:
        # Convert the graph to an undirected graph
        graph = graph.to_undirected()

    # Set the default settings for the graph visualization based on the number of nodes
    graph_settings, fig_size = get_graph_default_settings(len(graph.nodes))

    plt.figure(figsize=(fig_size, fig_size))
    node_colors = [hash(graph.nodes[node]["sheet"]) % 256 for node in graph.nodes]
    pos = nx.spring_layout(graph)  # layout for nodes

    # add legends for the colors
    node_colors, legend_patches = get_node_colors_and_legend(graph)

    nx.draw(
        graph,
        pos,
        node_color=node_colors,
        **graph_settings,
    )

    plt.legend(handles=legend_patches, title="Sheets", loc="upper left")

    filename = f"images/{file_path}.png"
    plt.savefig(filename)
    print(f"Graph visualization saved to {filename}")

    # open the image file in windows
    if sys.platform == "win32" and "--open-image" in sys.argv:
        import os

        os.system(f"start {filename}")



import pytest
from excel_parser import extract_references


# Helper function to assert references
def assert_references(formula, expected_direct, expected_range, expected_deps):
    direct_references, range_references, deps = extract_references(formula)
    assert (
        direct_references == expected_direct
    ), f"Expected {expected_direct}, but got {direct_references}"
    assert (
        range_references == expected_range
    ), f"Expected {expected_range}, but got {range_references}"
    assert deps == expected_deps, f"Expected {expected_deps}, but got {deps}"


@pytest.mark.parametrize(
    "formula, expected_direct, expected_range, expected_deps",
    [
        # Test for simple references like B4, A5
        ("=B4+A5", ["B4", "A5"], [], {}),
        # Test for local range references like A2:A11
        ("=SUM(A2:A4)", [], ["A2:A4"], {"A2": "A2:A4", "A3": "A2:A4", "A4": "A2:A4"}),
        # Test for simple absolute references like $A$1, $B$2
        ("=$A$1+$B$2", ["A1", "B2"], [], {}),
        # Test for sheet qualified absolute references like Sheet2!$A$1, Sheet2!$B$2
        ("=Sheet2!$A$1+Sheet2!$B$2", ["Sheet2!A1", "Sheet2!B2"], [], {}),
    ],
)
def test_references(formula, expected_direct, expected_range, expected_deps):
    """
    Test various cell and range references.
    """
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for sheet-qualified absolute range references like Sheet2!$A$1:$A$10
def test_sheet_qualified_absolute_range_references():
    """
    Test for sheet-qualified absolute range references like Sheet2!$A$2:$A$5.
    """
    formula = "=SUM(Sheet2!$A$2:$A$5)"
    expected_direct = []
    expected_range = ["Sheet2!A2:A5"]
    expected_deps = {
        "Sheet2!A2": "Sheet2!A2:A5",
        "Sheet2!A3": "Sheet2!A2:A5",
        "Sheet2!A4": "Sheet2!A2:A5",
        "Sheet2!A5": "Sheet2!A2:A5",
    }
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for sheet-qualified cell like Sheet2!C5
def test_sheet_qualified_reference():
    """
    Test for sheet-qualified cell like Sheet2!C5.
    """
    formula = "=Sheet2!C5"
    expected_direct = ["Sheet2!C5"]
    expected_range = []
    expected_deps = {}
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for expanded range in dependencies
def test_expanded_range_in_dependencies():
    """
    Test for expanded range in dependencies like A1:A3.
    """
    formula = "=SUM(A1:A3)"
    expected_direct = []
    expected_range = ["A1:A3"]
    expected_deps = {"A1": "A1:A3", "A2": "A1:A3", "A3": "A1:A3"}
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for no direct but only range references
def test_no_direct_but_only_range_references():
    """
    Test for no direct references but only range references like A1:A3.
    """
    formula = "=SUM(A1:A3)"
    expected_direct = []
    expected_range = ["A1:A3"]
    expected_deps = {"A1": "A1:A3", "A2": "A1:A3", "A3": "A1:A3"}
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for two ranges
def test_two_ranges():
    """
    Test for two ranges like A1:A3 and B1:B3.
    """
    formula = "=SUM(A1:A3) + SUM(B1:B3)"
    expected_direct = []
    expected_range = ["A1:A3", "B1:B3"]
    expected_deps = {
        "A1": "A1:A3",
        "A2": "A1:A3",
        "A3": "A1:A3",
        "B1": "B1:B3",
        "B2": "B1:B3",
        "B3": "B1:B3",
    }
    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for sheet-qualified range like Sheet2!A1:B10
def test_sheet_qualified_range():
    """
    Test for sheet-qualified range like Sheet2!A1:B3.
    """
    formula = "=SUM(Sheet2!A1:B3)"
    expected_direct = []
    expected_range = ["Sheet2!A1:B3"]

    expected_deps = {
        "Sheet2!A1": "Sheet2!A1:B3",
        "Sheet2!A2": "Sheet2!A1:B3",
        "Sheet2!A3": "Sheet2!A1:B3",
        "Sheet2!B1": "Sheet2!A1:B3",
        "Sheet2!B2": "Sheet2!A1:B3",
        "Sheet2!B3": "Sheet2!A1:B3",
    }

    assert_references(formula, expected_direct, expected_range, expected_deps)


# Test for mixed references with both local and sheet-qualified cells
def test_mixed_references():
    """
    Test for mixed references with both local and sheet-qualified cells.
    """
    formula = "=SUM(Sheet2!A1:B3, A5) + Sheet2!C5 + B6"
    expected_direct = ["A5", "Sheet2!C5", "B6"]
    expected_range = ["Sheet2!A1:B3"]
    expected_deps = {
        "Sheet2!A1": "Sheet2!A1:B3",
        "Sheet2!A2": "Sheet2!A1:B3",
        "Sheet2!A3": "Sheet2!A1:B3",
        "Sheet2!B1": "Sheet2!A1:B3",
        "Sheet2!B2": "Sheet2!A1:B3",
        "Sheet2!B3": "Sheet2!A1:B3",
    }
    assert_references(formula, expected_direct, expected_range, expected_deps)
